# -*- coding: utf-8 -*-
"""
unittest for WebTracker

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0 License, see LICENSE for more details.
"""

from datetime import datetime
import requests

import libs.webtracker
import dns.resolver

from libs.webtracker import WebTracker


class MockRequestResult:
    def __init__(self, code):
        self.status_code = code

    def close(self):
        pass

    @property
    def text(self):
        return '''
        <html>
        <head>
        <title>  Welcome</title>
        </head>
        <body>
        </body>
        </html>
        '''


def test_webtracker_allpaths(mocker):
    """
    Test all paths in WebTracker implementation by providing the following response.

    https://aiven.io --> 200 (ok case)
    https://contentmatched.com --> 200 (Page content matched, ok case)
    https://contentdoesntmatch.com --> 200 (Page content not expected)
    https://www.expireddns.com ---> 403 (none 200 case)
    https://notexistingdns.com --> Exception dns.resolver.NXDOMAIN case
    https://timeout.com --> timeout error (Exception requests.exceptions.ConnectTimeout)
    https://anyotherexception.com --> Any other exception

    :param mocker: pytest-mock fixturer
    """
    def mock_dns_resolver(domain, qtype):
        """
        A mock DNS resolver, which either return anything normally or raise dns.resolver.NXDOMAIN exception.
        For tests, it returns as follows.

        notexistingdns.com --> Exception dns.resolver.NXDOMAIN case
        Anything else --> '8.8.8.8'

        :param domain: Domain name
        :param qtype: query string, ignored.
        :return:
        """
        if domain == 'notexistingdns.com':
            raise dns.resolver.NXDOMAIN()
        else:
            return '8.8.8.8'

    def mock_requests_get(url, timeout):
        """
        A mock request get method which returns different response for different URLs.
        Basically it covers all test paths.
        Among the return,
        https://aiven.io --> 200 (ok case)
        https://contentmatched.com --> 200 (Page content matched, ok case)
        https://contentdoesntmatch.com --> 200 (Page content not expected)
        https://non200status.com --> 403 (non 200 case)
        https://www.sslfailure.com --->  Exception requests.exceptions.SSLError
        https://timeout.com --> timeout error (Exception requests.exceptions.ConnectTimeout)
        https://anyotherexception.com --> Any other exception

        :param url: request url
        :param timeout: timeout in seconds
        :return:
        """
        if url in ['https://aiven.io', 'https://contentdoesntmatch.com', 'https://contentmatched.com']:
            return MockRequestResult(200)
        elif url == 'https://non200status.com':
            return MockRequestResult(403)
        elif url == 'https://www.sslfailure.com':
            raise requests.exceptions.SSLError()
        elif url == 'https://timeout.com':
            raise requests.exceptions.ConnectTimeout()
        elif url == 'https://anyotherexception.com':
            raise Exception('Any Exception')

    reg_exp_nomatch = '<title>\\s*Bad.*\\s*</title>'
    reg_exp_match = '<title>\\s*Welcome.*\\s*</title>'
    websites = {'https://aiven.io': '', 'https://non200status.com': '', 'https://www.sslfailure.com': '',
                'https://notexistingdns.com': '', 'https://timeout.com': '',
                'https://anyotherexception.com': '', 'https://contentdoesntmatch.com': reg_exp_nomatch,
                'https://contentmatched.com': reg_exp_match}

    mocker.patch('libs.webtracker.dns.resolver.resolve', mock_dns_resolver)
    mocker.patch('libs.webtracker.requests.get', mock_requests_get)

    test_from = 'Sydney'
    tracker = WebTracker(test_from, websites)
    for webstatus in tracker.status():
        url = webstatus['url']
        assert webstatus['from'] == test_from.lower()
        currenttime = datetime.utcnow().timestamp()
        assert webstatus['timestamp'] <= currenttime  # Since it is mocked, they could be the same
        assert webstatus['timestamp'] > currenttime - 1

        if url in ['https://aiven.io', 'https://non200status.com', 'https://www.sslfailure.com',
                   'https://contentdoesntmatch.com', 'https://contentmatched.com']:
            assert webstatus['status'] == 'responsive'
        else:
            assert webstatus['status'] == 'unresponsive'
        if url in ['https://aiven.io', 'https://contentmatched.com']:
            assert webstatus['phrase'] == 'ok'
        elif url in ['https://non200status.com']:
            assert webstatus['phrase'] == tracker.status_mapping[403]
        elif url in ['https://www.sslfailure.com']:
            assert webstatus['phrase'] == 'ssl error'
        elif url in ['https://notexistingdns.com']:
            assert webstatus['phrase'] == 'domain not exist'
        elif url in ['https://timeout.com']:
            assert webstatus['phrase'] == 'connection timeout'
        elif url in ['https://anyotherexception.com']:
            assert webstatus['phrase'] == 'any exception'
        elif url in ['https://contentdoesntmatch.com']:
            assert webstatus['phrase'] == 'page content not expected'
            assert webstatus['detail'] == reg_exp_nomatch

        if url in ['https://aiven.io', 'https://non200status.com', 'https://www.sslfailure.com',
                   'https://timeout.com', 'https://anyotherexception.com',
                   'https://contentdoesntmatch.com', 'https://contentmatched.com']:
            assert webstatus['dns'] >= 0
        else:
            assert webstatus['dns'] is None

        if url in ['https://aiven.io', 'https://non200status.com',
                   'https://contentdoesntmatch.com', 'https://contentmatched.com']:
            assert webstatus['response'] >= 0
        else:
            assert webstatus['response'] is None



# -*- coding: utf-8 -*-
"""
libs.status
~~~~~~~~~~~~~~~

This module implements WebsiteStatus.

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0 License, see LICENSE for more details.
"""
from http import HTTPStatus
from datetime import datetime
from urllib.parse import urlparse
import msgpack


class WebsiteStatus(dict):
    """
    An abstract of Website status, which represents a corresponding status generated by WebTracker and provides
    APIs to create DB related schemas, insert and update operations through given DB connection instance.
    """

    # A cached dictionary tracking last status for every URL. The key is URL and value is of
    # (True/False, status record id), with True for healthy and False for unhealthy.
    last_url_status = {}

    def __init__(self, status_from, url, status, phrase, dns_time=None, response_time=None, detail='', offset=-1):
        """
        Constructor
        :param status_from: where this test status is from, always transferred to lower cases.
        :param url: Website URL, like https://aiven.io
        :param status: One of "responsive, "unresponsive"
        :param phrase: Status description. This includes all HTTP status code phrases like 'ok', 'forbidden', etc, and
        four additional phrases: 'domain not exist', 'ssl error', 'connection timeout' and 'Page content not expected'.
        :param dns_time: Time in ms taken to resolve DNS.
        :param response_time: Time in ms taken totally to load the Website page.
        :param detail: When phrase is 'Page content not expected', detail shows regular expression for this match.
        :param offset: offset of Kafka message in the corresponding topic
        """
        # Status' UTC timestamp
        timestamp = datetime.utcnow().timestamp()
        # Domain is used as the topic with dots replaced with underscores.
        self.topic = urlparse(url).netloc.replace('.', '_')

        super().__init__({'from': status_from.lower(), 'url': url.lower(), 'timestamp': timestamp,
                          'status': status.lower(), 'phrase': phrase.lower(), 'dns': dns_time,
                          'response': response_time, 'detail': detail, 'offset': offset})

    @classmethod
    def create_type_schema(cls, db):
        """
        Given a PostgreSQL connection, create enum type for response status and phrases.
        The intention is to save space by using enum type instead of strings directly in DB.
        :param db: A PostgresSQL DB connection instance.
        :return:
        """
        response_status_type_sql = "CREATE TYPE response_status AS ENUM ('responsive', 'unresponsive');"
        # A complete phrases include all HTTP status codes and four additional ones.
        phrases = [getattr(x, 'phrase').lower() for x in HTTPStatus] + ['domain not exist', 'ssl error',
                                                                        'connection timeout',
                                                                        'page content not expected']
        phrases = tuple(phrases)
        phrase_type_sql = f'CREATE TYPE phrase_status AS ENUM {phrases};'

        with db.cursor() as curs:
            for typename, sql in [('response_status', response_status_type_sql), ('phrase_status', phrase_type_sql)]:
                # Firstly check whether this type exits
                curs.execute(f"select 1 from pg_type where typname = '{typename}'")
                if curs.fetchone():  # Defined already, do nothing
                    continue
                curs.execute(sql)
        db.commit()

    @classmethod
    def create_table_schema(cls, topic, db):
        """
        Given a DB connection, create a corresponding PostgreSQL table for the specific topic.
        The table name is in the format of web_activity_<topic>, which dots are replaced with underscores.
        :param topic: The specific topic to be created for
        :param db: A PostgresSQL DB connection instance.
        """
        topic = topic.replace('.', '_')
        table_sql = f'''
        CREATE TABLE IF NOT EXISTS web_activity_{topic} (
        id SERIAL PRIMARY KEY,
        created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
        topic_offset BIGINT DEFAULT -1,
        test_from varchar(32) NOT NULL,
        url varchar(256) NOT NULL,
        event_time BIGINT,
        status response_status,
        phrase phrase_status,
        dns integer DEFAULT -1,
        response integer DEFAULT -1,
        detail varchar(128)
        );
        '''
        index_sql = f'''
        CREATE INDEX IF NOT EXISTS web_activity_url_index_{topic} ON web_activity_{topic} (url);
        '''
        with db.cursor() as curs:
            curs.execute(table_sql)
            curs.execute(index_sql)
        db.commit()

    @classmethod
    def get_last_offset(cls, topic, db):
        """
        Get the last offset received for the specific topic
        :param topic: the topic to get offset for
        :param db: DB connection
        :return: The last offset received from Kafka. -1 if nothing received yet
        """
        topic = topic.replace('.', '_')
        sql = f'''
        SELECT topic_offset from web_activity_{topic} order by id DESC LIMIT 1;
        '''
        with db.cursor() as curs:
            curs.execute(sql)
            r = curs.fetchone()
            if r:
                return r[0]
            else:
                return -1

    def healthy(self):
        """
        Return whether it is a healthy status. A status is only healthy when:
        1. status is 'responsive'
        2. phrase is 'ok'
        :return: True or False
        """
        return self['status'] == 'responsive' and self['phrase'] == 'ok'

    def insert_status(self, db):
        """
        Insert this instance to DB as a new row. Its side effect is to update WebsiteStatus::last_url_status cache.
        :param db: DB connection
        """
        sql = f'''INSERT INTO web_activity_{self.topic} (topic_offset, test_from, url,
        event_time, status, phrase, dns, response, detail) VALUES (
        {self["offset"]}, '{self["from"]}', '{self["url"]}', {int(self["timestamp"])},
        '{self["status"]}', '{self["phrase"]}', {self["dns"] if self["dns"] else -1},
        {self["response"] if self["response"] else -1}, '{self["detail"]}'
        );'''
        last_id_sql = 'SELECT LASTVAL();'
        with db.cursor() as curs:
            curs.execute(sql)
            curs.execute(last_id_sql)
            r = curs.fetchone()
            if r:  # This should always meet, just to make sure.
                self.last_url_status[self['url']] = (self.healthy(), r[0])

        db.commit()

    def insert_status_smart(self, db):
        """
        Instead of inserting a new status always, this API tries to update the last healthy status if the healthy
        status record is the last one. Otherwise, insert_status is still called.
        This is superb helpful to reduce many rows of similar healthy records.
        """
        if self['url'] not in self.last_url_status:
            no_record = False

            last_status_sql = f'''
            SELECT id, status, phrase from web_activity_{self.topic} WHERE url = '{self["url"]}'
            ORDER BY id DESC LIMIT 1;
            '''
            with db.cursor() as curs:
                curs.execute(last_status_sql)
                r = curs.fetchone()
                if r:
                    rid, status, phrase = r
                    if status == 'responsive' or phrase == 'ok':
                        self.last_url_status[self['url']] = (True, rid)
                    else:
                        self.last_url_status[self['url']] = (False, rid)
                else:
                    no_record = True
            db.commit()

            if no_record:  # No record found which means no healthy record at all. Do insert instead
                self.insert_status(db)
                return

        if self.last_url_status[f'{self["url"]}'][0] and self.healthy():
            # Last status record is healthy and this new status is healthy
            # Update the last healthy record with latest created_at, topic_offset, event_time
            sql = f'''
            UPDATE web_activity_{self.topic} SET (created_at, topic_offset, event_time) = (CURRENT_TIMESTAMP,
            {self['offset']}, {self['timestamp']}) WHERE id = {self.last_url_status[self['url']][1]};
            '''
            with db.cursor() as curs:
                if self.last_url_status[self['url']][0]:
                    curs.execute(sql)
            db.commit()
        else:  # Otherwise, insert a new one
            self.insert_status(db)

    def serialize(self):
        """
        Return a serialized bytes
        :return: bytes
        """
        return msgpack.packb(self, use_bin_type=True)

    @classmethod
    def deserialize(cls, raw_bytes):
        """
        Deserialize raw bytes into WebsiteStatus instance
        :param raw_bytes: bytes content
        :return: WebsiteStatus instance
        """
        dictv = msgpack.unpackb(raw_bytes, raw=False)
        reborn = WebsiteStatus(dictv['from'], dictv['url'], dictv['status'], dictv['phrase'], dictv['dns'],
                               dictv['response'], dictv['detail'], dictv['offset'])
        # timestamp has to be reset to the original value since reborn comes with a newly created timestamp
        reborn['timestamp'] = dictv['timestamp']
        return reborn



# -*- coding: utf-8 -*-
"""
Logging implementation.
This is a logging module which enables you local logging or remote logging to a Kafka Server under a specific topic.

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0, see LICENSE for more details.
"""

import logging

from kafka.admin import NewTopic
from kafka import errors as KafkaErrors


class KafkaLoggingHandler(logging.Handler):
    """
    A simple handler which emits records to Kafka server
    """

    def __init__(self, kafka_producer, topic):
        """
        Constructor
        :param kafka_producer: A Kafka Producer instance
        :param topic: Kafka topic
        """
        super().__init__()
        self.topic = topic
        self.producer = kafka_producer

    def emit(self, record):
        msg = self.format(record)
        self.producer.send(self.topic, bytes(msg, 'utf-8'))

    def close(self):
        self.producer.close()
        super().close()


class Logger:
    """
    A logger which either log to local sys.stderr or remote Kafka server.
    """

    def __init__(self, name, kafka_producer=None, kafka_admin=None):
        """
        Constructor
        :param name: Logger name
        :param kafka_server: A Kafka Producer instance if remote Kafka logging is wanted instead.
        :param kafka_admin: A Kafka KafkaAdminClient for creating topic
        """
        self.log = logging.getLogger(name)
        topic = f'logging-{name}'
        if kafka_producer:
            handler = KafkaLoggingHandler(kafka_producer, topic)
            if kafka_admin:
                ktopic = NewTopic(name=topic, num_partitions=1, replication_factor=2)
                try:
                    kafka_admin.create_topics([ktopic], validate_only=False)
                except KafkaErrors.TopicAlreadyExistsError:
                    pass
        else:
            handler = logging.StreamHandler()
        f = logging.Formatter('%(asctime)s - %(name)s - %(levelname)-8s - %(message)s')
        handler.setFormatter(f)
        self.log.addHandler(handler)
        self.log.setLevel(logging.INFO)

    def __getattr__(self, name):
        """
        Redirect all methods to self.log which is logging.Logger instance.
        :param name: Method/Attribute name
        :return: Method/Attribute from self.log
        """
        return getattr(self.log, name)



# -*- coding: utf-8 -*-

"""
tracker.tracker
~~~~~~~~~~~~~~~

This module implements WebTracker which sends HTTP GET request to provided URLs and provides their response status.

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0 License, see LICENSE for more details.
"""

import time
import re
import requests
import copy
from urllib.parse import urlparse
from http import HTTPStatus

import dns.resolver

from .status import WebsiteStatus


class WebTracker:
    """
    Class to report Websites Availability.
    """

    def __init__(self, track_from, urls={}, timeout=10):
        """
        Constructor
        :param track_from: Where this Web Availability test is made, like 'sydney'.
        It would be converted to lower cases.
        :param urls: A dictionary with URL and its expected regexp pattern for the page,
        For example, {'https://aiven.io':'<svg\\s+.*', 'https://www.google.com':''} is:
        1. Detect 'https://aiven.io' status and checking whether the response page contains svg tag
        2. Detect 'https://www.goog.com' status
        :param timeout: Seconds to wait for the server to send data before giving up
        """
        self.track_from = track_from.lower()
        # Original dictionary with regexp string
        self.urls = urls
        # This new dictionary with one-off compiled regexp
        self._urls = copy.deepcopy(self.urls)
        for url, reg_exp in urls.items():
            if reg_exp.strip():
                self._urls[url] = re.compile(reg_exp.strip())
        self.timeout = timeout
        # A mapping between HTTP Status codes and their phrases.
        self.status_mapping = dict([(getattr(x, 'value'), getattr(x, 'phrase').lower()) for x in HTTPStatus])

    def status(self):
        """
        A generator to return status for all URLs at the specific period
        :return: A generator yielding WebsiteStatus objects
        """
        for url, re_exp in self._urls.items():
            status = 'unresponsive'
            dns_time = None
            response_time = None
            detail = None

            # We can resolve domain one-off in __init__ but decided not to do so
            # since this is not a time-critical task and it is handy to do in this way.
            domain = urlparse(url).netloc
            try:
                # Measure DNS query time firstly
                stime = time.time()
                # We are measuring DNS response in a rough way through dns.resolver module
                dns.resolver.resolve(domain, 'A')
                etime = time.time()
                # DNS response time in ms
                dns_time = int((etime - stime) * 1000)
                stime = time.time()
                r = requests.get(url, timeout=self.timeout)
                r.close()
                etime = time.time()
                # Total Response time in ms
                response_time = int((etime - stime) * 1000)
                status = 'responsive'
                phrase = self.status_mapping[r.status_code]
                if re_exp:
                    if not re_exp.findall(r.text):
                        phrase = 'Page content not expected'
                        detail = f'{self.urls[url]}'
            except dns.resolver.NXDOMAIN:
                phrase = 'domain not exist'
            except requests.exceptions.SSLError:
                status = 'responsive'
                phrase = 'ssl error'
            except requests.exceptions.ConnectTimeout:
                phrase = 'connection timeout'
            except Exception as e:  # This should never happen, but just in case.
                phrase = str(e).lower()
            yield WebsiteStatus(self.track_from, url, status, phrase, dns_time, response_time, detail)



# -*- coding: utf-8 -*-
"""
unittest for WebsiteStatus

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0 License, see LICENSE for more details.
"""
from contextlib import contextmanager
from http import HTTPStatus
from random import randint
import msgpack
from libs.status import WebsiteStatus


# A complete phrases include all HTTP status codes and four additional.
phrases = [getattr(x, 'phrase').lower() for x in HTTPStatus] + ['domain not exist', 'ssl error',
                                                                'connection timeout', 'page content not expected']


class MockDB:
    """
    A mock DB Connection to verify SQl statement.
    """
    def __init__(self, something_to_fetch=[]):
        self.sqls = []
        self.curs = None
        self.something_to_fetch = something_to_fetch

    class MockCursor:
        def __init__(self, something_to_fetch=[]):
            self.sqls = []
            self.something_to_fetch = something_to_fetch

        def execute(self, sql):
            self.sqls.append(sql)

        def fetchone(self):
            try:
                return self.something_to_fetch.pop(0)
            except IndexError:
                return None

    @contextmanager
    def cursor(self):
        if self.curs is None:
            self.curs = MockDB.MockCursor(self.something_to_fetch)
        yield self.curs
        self.sqls = self.curs.sqls

    def commit(self):
        pass


def sqls_compare(sql_list1, sql_list2):
    """
    To compare two sql statements lists. Two sql lists are equal only when
    1. Contains the list has the same number of sql statements
    2. The tokens in every corresponding statement are the same in the order
    """
    sql1_tokens = [sql.split() for sql in sql_list1]
    sql2_tokens = [sql.split() for sql in sql_list2]
    return sql1_tokens == sql2_tokens


def test_type_schema():
    """
    Verify WebsiteStatus::create_type_schema
    """
    response_status_type_select_sql = "select 1 from pg_type where typname = 'response_status'"
    response_status_type_sql = "CREATE TYPE response_status AS ENUM ('responsive', 'unresponsive');"
    phrases_type_select_sql = "select 1 from pg_type where typname = 'phrase_status'"
    phrases_tuple = tuple(phrases)
    phrase_type_sql = f'CREATE TYPE phrase_status AS ENUM {phrases_tuple};'

    db = MockDB()
    WebsiteStatus.create_type_schema(db)
    # Verify 'CREATE TYPE' statement if TYPE phrase_status doesn't exit
    assert sqls_compare(db.sqls, [response_status_type_select_sql, response_status_type_sql,
                                  phrases_type_select_sql, phrase_type_sql])
    db = MockDB([[1], [1]])
    WebsiteStatus.create_type_schema(db)
    # Verify no 'CREATE TYPE' statement is executed if TYPE phrase_status exits
    assert sqls_compare(db.sqls, [response_status_type_select_sql, phrases_type_select_sql])


def test_table_schema():
    """
    Verify WebsiteStatus::create_table_schema
    """
    table_sql = '''
        CREATE TABLE IF NOT EXISTS web_activity_aiven_io (
        id SERIAL PRIMARY KEY,
        created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
        topic_offset BIGINT DEFAULT -1,
        test_from varchar(32) NOT NULL,
        url varchar(256) NOT NULL,
        event_time BIGINT,
        status response_status,
        phrase phrase_status,
        dns integer DEFAULT -1,
        response integer DEFAULT -1,
        detail varchar(128)
        );
        '''
    index_sql = '''
        CREATE INDEX IF NOT EXISTS web_activity_url_index_aiven_io ON web_activity_aiven_io (url);
        '''
    db = MockDB()
    WebsiteStatus.create_table_schema('aiven.io', db)
    assert sqls_compare(db.sqls, [table_sql, index_sql])


def test_get_last_offset():
    """
    Verify WebsiteStatus::get_last_offset by checking a correct SELECT SQL is invoked and a correct
    offset is returned.
    """
    sql = '''
        SELECT topic_offset from web_activity_aiven_io order by id DESC LIMIT 1;
        '''
    # Verify returning -1 when there is nothing
    db = MockDB()
    r = WebsiteStatus.get_last_offset('aiven.io', db)
    assert db.sqls == [sql]
    assert r == -1
    # Verify returning the correct number for an existing last offset
    rnum = randint(1, 0xFFFFFFFF)
    db = MockDB([[rnum]])
    r = WebsiteStatus.get_last_offset('aiven.io', db)
    assert sqls_compare(db.sqls, [sql])
    assert r == rnum


def test_healthy():
    """
    Verify WebsiteStatus::healthy
    """
    # All cases' verification
    for status in ['responsive', 'unresponsive']:
        for phrase in phrases:
            web_status = WebsiteStatus('Sydney', 'https://aiven.io', status, phrase, 10, 300)
            if status == 'responsive' and phrase == 'ok':
                assert web_status.healthy() is True
            else:
                assert web_status.healthy() is False


def test_insert_status():
    """
    Verify WebsiteStatus::insert_status by checking a correct INSERT SQL is submitted to DB.
    """
    last_id_sql = 'SELECT LASTVAL();'

    # All cases' verification
    for status in ['responsive', 'unresponsive']:
        for phrase in phrases:
            web_status = WebsiteStatus('Sydney', 'https://aiven.io', status, phrase, 10, 300)
            sql = f'''INSERT INTO web_activity_aiven_io (topic_offset, test_from, url,
            event_time, status, phrase, dns, response, detail) VALUES (
            {web_status["offset"]}, '{web_status["from"]}', '{web_status["url"]}', {int(web_status["timestamp"])},
            '{web_status["status"]}', '{web_status["phrase"]}', {web_status["dns"] if web_status["dns"] else -1},
            {web_status["response"] if web_status["response"] else -1},
            '{web_status["detail"]}'
            );'''
            # A random number to emulate last insert ID
            rnum = randint(1, 0xFFFFFFFF)
            db = MockDB([[rnum]])
            web_status.insert_status(db)
            # Verify executed statements
            assert sqls_compare(db.sqls, [sql, last_id_sql])
            # Verify WebsiteStatus::insert_status updated
            if status == 'responsive' and phrase == 'ok':
                assert WebsiteStatus.last_url_status['https://aiven.io'] == (True, rnum)
            else:
                assert WebsiteStatus.last_url_status['https://aiven.io'] == (False, rnum)


def test_insert_status_smart():
    """
    Test WebsiteStatus::insert_status_smart
    """
    last_status_sql = '''
        SELECT id, status, phrase from web_activity_aiven_io WHERE url = 'https://aiven.io'
        ORDER BY id DESC LIMIT 1;
    '''
    insert_sql = '''INSERT INTO web_activity_aiven_io (topic_offset, test_from, url,
    event_time, status, phrase, dns, response, detail) VALUES (
    %d, 'sydney', 'https://aiven.io', %d, '%s', '%s', %d, %d, '%s'
    );'''
    last_id_sql = 'SELECT LASTVAL();'
    # Case 1: Verify a new status record is inserted when table is empty
    WebsiteStatus.last_url_status = {}
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300, '', 200)
    rnum = randint(1, 0xFFFFFFFF)
    db = MockDB([None, [rnum]])
    web_status.insert_status_smart(db)
    insert_sql_value = insert_sql % (web_status['offset'], web_status['timestamp'], 'responsive', 'ok', 10, 300, '')
    # Verify SQLs
    assert sqls_compare(db.sqls, [last_status_sql, insert_sql_value, last_id_sql])
    # Verify WebsiteStatus.last_url_status cache's content
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (True, rnum)

    # Case 2: Verify a healthy record updates the existing record if the table is not empty
    WebsiteStatus.last_url_status = {}
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300, '', 200)
    rnum1 = randint(1, 0xFFFFFFFF)
    rnum2 = randint(1, 0xFFFFFFFF)
    db = MockDB([[rnum1, 'responsive', 'ok'], [rnum2]])
    update_sql = f'''
    UPDATE web_activity_aiven_io SET (created_at, topic_offset, event_time) = (CURRENT_TIMESTAMP,
    {web_status['offset']}, {web_status['timestamp']}) WHERE id = {rnum1};
    '''
    web_status.insert_status_smart(db)
    # Verify SQLs
    assert sqls_compare(db.sqls, [last_status_sql, update_sql])
    # Verify WebsiteStatus.last_url_status contains lasting healthy record from table
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (True, rnum1)

    # Case 3: Verify a healthy record updates the existing record if it is healthy too
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300, '', 201)
    db = MockDB()
    update_sql = f'''
    UPDATE web_activity_aiven_io SET (created_at, topic_offset, event_time) = (CURRENT_TIMESTAMP,
    {web_status['offset']}, {web_status['timestamp']}) WHERE id = {WebsiteStatus.last_url_status[web_status['url']][1]};
    '''
    # Insert another record, which is supposed to update the last one only
    web_status.insert_status_smart(db)
    assert sqls_compare(db.sqls, [update_sql])
    # Verify WebsiteStatus.last_url_status cache keeps unchanged since it is an update
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (True, rnum1)

    # Case 4: Verify an unhealthy record is inserted when the last record is healthy
    rnum = randint(1, 0xFFFFFFFF)
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'unresponsive', 'connection timeout', 10, 300, '', rnum)
    db = MockDB([[rnum]])
    web_status.insert_status_smart(db)
    insert_sql_value = insert_sql % (web_status['offset'], web_status['timestamp'], web_status['status'],
                                     web_status['phrase'], 10, 300, '')
    assert sqls_compare(db.sqls, [insert_sql_value, last_id_sql])
    # Verify WebsiteStatus.last_url_status cache is updated
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (False, rnum)

    # Case 5: Verify a healthy record is inserted if the last record is unhealthy
    rnum = randint(1, 0xFFFFFFFF)
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ok', 10, 300, '', rnum)
    db = MockDB([[rnum]])
    web_status.insert_status_smart(db)
    insert_sql_value = insert_sql % (web_status['offset'], web_status['timestamp'], web_status['status'],
                                     web_status['phrase'], 10, 300, '')
    assert sqls_compare(db.sqls, [insert_sql_value, last_id_sql])
    # Verify WebsiteStatus.last_url_status cache is updated
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (True, rnum)

    # Case 6: Verify an unhealthy status record is inserted when the last record is unhealthy too
    rnum = randint(1, 0xFFFFFFFF)
    web_status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'ssl error', 10, 300, '', rnum)
    db = MockDB([[rnum]])
    web_status.insert_status_smart(db)
    insert_sql_value = insert_sql % (web_status['offset'], web_status['timestamp'], web_status['status'],
                                     web_status['phrase'], 10, 300, '')
    assert sqls_compare(db.sqls, [insert_sql_value, last_id_sql])
    # Verify WebsiteStatus.last_url_status cache is updated
    assert WebsiteStatus.last_url_status['https://aiven.io'] == (False, rnum)


def test_serialize():
    """
    Test WebsiteStatus::serialize
    """
    # All cases' verification
    for status in ['responsive', 'unresponsive']:
        for phrase in phrases:
            if phrase == 'page content not expected':
                detail = '<svg\\s+'
            else:
                detail = ''
            webstatus = WebsiteStatus('Sydney', 'https://aiven.io', status, phrase, 10, 300, detail, 100)
            s1 = webstatus.serialize()
            s2 = {'from': 'sydney', 'url': 'https://aiven.io', 'timestamp': webstatus['timestamp'],
                  'status': status, 'phrase': phrase, 'dns': 10, 'response': 300, 'detail': detail, 'offset': 100}
            # webstatus is a dictionary
            assert s1 == msgpack.packb(s2, use_bin_type=True)


def test_deserialize():
    """
    Test WebsiteStatus::deserialize
    """
    # This raw is a msgpack.packb result from WebsiteStatus('Sydney', 'https://aiven.io', 'responsive',
    #                                                       'forbidden', 10, 300, '', 15)
    raw = b'\x89\xa4from\xa6sydney\xa3url\xb0https://aiven.io\xa9timestamp\xcbA\xd7\xf9L\x02d.\xa1\xa6status\xaa'\
          b'responsive\xa6phrase\xa9forbidden\xa3dns\n\xa8response\xcd\x01,\xa6detail\xa0\xa6offset\x0f'
    # Create the original WebsiteStatus object for this serialized bytes.
    status = WebsiteStatus('Sydney', 'https://aiven.io', 'responsive', 'forbidden', 10, 300, '', 15)
    # Reset timestamp to expected value since timestamp is created lively for every instance.
    # Therefore, it has to be reset to the value for raw
    status['timestamp'] = 1608855561.565346
    # Verify the deserialized one equals to the original one
    assert WebsiteStatus.deserialize(raw) == status



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Main file of Recorder component. It tries to pull messages from the specific Kafka Server for its interested domains
and insert them into the specific tables on specific PostgreSQL server.

usage: recorder.py [-h] [-d DOMAIN] [-p POSTGRES] [-g PGCA] [-k KAFKA] [-c CA] [-e CERT] [-y KEY]

Websites Availability Recorder

optional arguments:
  -h, --help            show this help message and exit
  -d DOMAIN, --domain DOMAIN
                        A list of domains to be watched, separated by comma. Like 'aiven.io,google.com'.
                        It will be overridden by the environment variable DOMAIN.
  -p POSTGRES, --postgres POSTGRES
                        PostgresSQL server URI. A sample like
                        postgres://avnadmin:tvezg@pg-9971643.aivencloud.com:20274/defaultdb?sslmode=require
                        It can be overridden by the environment variable POSTGRES.
  -g PGCA, --pgca PGCA  PostgresSQL server CA file. It can be overridden by the environment variable PGCA.
  -k KAFKA, --kafka KAFKA
                        Kafka bootstrap server. It can be overridden by the environment variable KAFKA.
  -c CA, --ca CA        Kafka server CA(Certificate Authority) file path.
                        The CA content's can be overridden by the environment variable CA.
  -e CERT, --cert CERT  This client certificate file path.
                        The certificate's content can be overridden by the environment variable CERT.
  -y KEY, --key KEY     This client private key file path.
                        The private key's content can be overridden by the environment variable KEY.
  -s, --smart           Smart mode. In this mode, continuous healthy records are merged into one with event time
                        updated. It saves a lot of DB space since most of records are healthy. It can be overridden by
                        the environment variable SMART.

:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0 License, see LICENSE for more details.
"""

import os
import argparse

from kafka import KafkaConsumer, KafkaProducer
from kafka.admin import KafkaAdminClient
from kafka.structs import TopicPartition
import psycopg2
from libs.status import WebsiteStatus
from libs.log import Logger


def main():
    parser = argparse.ArgumentParser(description='Websites Availability Recorder')
    parser.add_argument('-d', '--domain', help='''
    A list of domains to be watched, separated by comma. Like 'aiven.io,google.com'. It will be overridden by the
    environment variable DOMAIN.
    ''')
    parser.add_argument('-p', '--postgres', help='''
    PostgresSQL server URI.
    A sample like postgres://avnadmin:rbb8en66jn7tvezg@pg-9971643.aivencloud.com:20274/defaultdb?sslmode=require
    It can be overridden by the environment variable POSTGRES.
    ''')
    parser.add_argument('-g', '--pgca', help='''
    PostgresSQL server CA file. It can be overridden by the environment variable PGCA.
    ''')
    parser.add_argument('-k', '--kafka', help='''
    Kafka bootstrap server. It can be overridden by the environment variable KAFKA.
    ''')
    parser.add_argument('-c', '--ca', help='''
    Kafka server CA(Certificate Authority) file path. The CA content's can be overridden by the environment
    variable CA.
    ''')
    parser.add_argument('-e', '--cert', help='''
    This client certificate file path. The certificate's content can be overridden by the environment variable CERT.
    ''')
    parser.add_argument('-y', '--key', help='''
    This client private key file path. The private key's content can be overridden by the environment variable KEY.
    ''')
    parser.add_argument('-s', '--smart', action='store_true', default=False, help='''
    Smart mode. In this mode, continuous healthy records are merged into one with event time updated. It saves a lot
    of DB space since most of records are healthy. It can be overridden by the environment variable SMART.
    ''')

    args = parser.parse_args()

    domains = os.getenv('DOMAIN') if os.getenv('DOMAIN') else args.domain
    domains = domains.split(',')
    pguri = os.getenv('POSTGRES') if os.getenv('POSTGRES') else args.postgres
    pgca = os.getenv('PGCA') if os.getenv('PGCA') else args.pgca
    kafka = os.getenv('KAFKA') if os.getenv('KAFKA') else args.kafka
    ca = os.getenv('CA') if os.getenv('CA') else args.ca
    cert = os.getenv('CERT') if os.getenv('CERT') else args.cert
    key = os.getenv('KEY') if os.getenv('KEY') else args.key
    smart = os.getenv('SMART') if os.getenv('SMART') else args.smart

    conn = None
    try:
        conn = psycopg2.connect(pguri, sslrootcert=pgca)
    except Exception as e:
        print(f'Could not connect postgreSQL - {e}')
        exit(-1)

    consumer = KafkaConsumer(
        auto_offset_reset="earliest",
        bootstrap_servers=kafka,
        value_deserializer=WebsiteStatus.deserialize,
        enable_auto_commit=True,
        security_protocol="SSL",
        ssl_cafile=ca,
        ssl_certfile=cert,
        ssl_keyfile=key
    )

    if os.getenv('PRODUCTION'):
        # Create Kafka producer for logging
        producer = KafkaProducer(
            bootstrap_servers=kafka,
            security_protocol="SSL",
            ssl_cafile=ca,
            ssl_certfile=cert,
            ssl_keyfile=key
        )
        # Create Kafka admin client
        admin = KafkaAdminClient(
            bootstrap_servers=kafka,
            security_protocol="SSL",
            ssl_cafile=ca,
            ssl_certfile=cert,
            ssl_keyfile=key
        )
        logger = Logger('recorder', producer, admin)
        # admin client is used once for creating logging topic
        admin.close()
    else:
        logger = Logger('recorder')

    # Add related type schema if necessary
    WebsiteStatus.create_type_schema(conn)

    tps = {}
    for domain in domains:
        # Create table schema for this domain
        WebsiteStatus.create_table_schema(domain, conn)
        tp = TopicPartition(f'webactiviity-{domain}', 0)
        tps[domain] = tp

    # Assign topics
    consumer.assign(tps.values())

    # Now seek to last positions
    for domain, tp in tps.items():
        # retrieve next position for the topic/domain
        consumer.seek(tp, WebsiteStatus.get_last_offset(domain, conn) + 1)
        logger.info(f'Starting position for {domain}: {consumer.position(tp)}')

    count = 0
    while True:
        msg_pack = consumer.poll(timeout_ms=2000)
        for tp, messages in msg_pack.items():
            for message in messages:
                status = message.value
                status['offset'] = message.offset
                logger.info("count: %d - %s:%d:%d: key=%s value=%s" % (count, tp.topic, tp.partition,
                                                                       message.offset, message.key,
                                                                       message.value))
                if smart:
                    status.insert_status_smart(conn)
                else:
                    status.insert_status(conn)
                count += 1


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Main file of Tracker component. It tries to access each specified URL, generate status and push it to a specified
Kafka server under its associated topic.

usage: tracker.py [-h] [-l LOCATION] [-w WEBSITES] [-t TIMEOUT] [-p PERIOD] [-k KAFKA] [-c CA] [-e CERT] [-y KEY]

Websites Availability Tracker

optional arguments:
  -h, --help            show this help message and exit
  -l LOCATION, --location LOCATION
                        The location to tell where this tracker runs from. This is important to tell the availability
                        from a specific location.
  -w WEBSITES, --websites WEBSITES
                        A list of websites URLs to be watched with their optional regular expression to match for the
                        returned page. Separated by comma.
                        Note: URL and its regular expression is separated with spaces.
                        Like 'https://aiven.io <svg\\s+,https://www.google.com',
                        which means:
                        1. https://aiven.io will be watched and loaded page is expected to match regexp '<svg\\s+'
                        2. https://www.google.com will be watched
                        It can be overridden by the environment variable WEBSITES.
  -t TIMEOUT, --timeout TIMEOUT
                        The timeout in seconds to determine connection timeout. Default is 10 seconds.
                        It can be overridden by the environment variable TIMEOUT.
  -p PERIOD, --period PERIOD
                        Period time in seconds at least which all Websites Availability will be monitored.
                        Default is 300 seconds. It can be overridden by environment variable PERIOD.
  -k KAFKA, --kafka KAFKA
                        Kafka bootstrap server. It can be overridden by the environment variable KAFKA.
  -c CA, --ca CA        Kafka server CA(Certificate Authority) file path.
                        The CA content's can be overridden by the environment variable CA.
  -e CERT, --cert CERT  This client certificate file path.
                        The certificate's content can be overridden by the environment variable CERT.
  -y KEY, --key KEY     This client private key file path.
                        The private key's content can be overridden by the environment variable KEY.


:copyright: (c) 2020 by Liang Hou.
:license: Apache-2.0  License, see LICENSE for more details.
"""

import os
import time
from urllib.parse import urlparse
import argparse

from kafka import KafkaProducer
from kafka import errors as KafkaErrors
from kafka.admin import KafkaAdminClient, NewTopic
from libs.webtracker import WebTracker
from libs.log import Logger


def main():
    parser = argparse.ArgumentParser(description='Websites Availability Tracker')
    parser.add_argument('-l', '--location', help='''
    The location to tell where this tracker runs from. This is important to tell the availability from a specific
    location.''')
    parser.add_argument('-w', '--websites', help='''
    A list of websites URLs to be watched with their optional regular expression to match for the returned page.
    Separated by comma. Note: URL and its regular expression is separated with spaces.
    Like 'https://aiven.io <svg\\s+,https://www.google.com', which means:
    1. https://aiven.io will be watched and loaded page is expected to match regexp '<svg\\s+'
    2. https://www.google.com will be watched
    It can be overridden by the environment variable WEBSITES.
    ''')
    parser.add_argument('-t', '--timeout', type=int, default=5, help='''
    The timeout in seconds to determine connection timeout. Default is 10 seconds. It can be overridden by the
    environment variable TIMEOUT.
    ''')
    parser.add_argument('-p', '--period', type=int, default=300, help='''
    Period time in seconds at least which all Websites Availability will be monitored. Default is 300 seconds.
    It can be overridden by environment variable PERIOD.
    ''')
    parser.add_argument('-k', '--kafka', help='''
    Kafka bootstrap server. It can be overridden by the environment variable KAFKA.
    ''')
    parser.add_argument('-c', '--ca', help='''
    Kafka server CA(Certificate Authority) file path. It can be overridden by the environment variable CA.
    ''')
    parser.add_argument('-e', '--cert', help='''
    This client certificate file path. It can be overridden by the environment variable CERT.
    ''')
    parser.add_argument('-y', '--key', help='''
    This client private key file path. It can be overridden by the environment variable KEY.
    ''')

    args = parser.parse_args()

    location = os.getenv('LOCATION') if os.getenv('LOCATION') else args.location
    websites = os.getenv('WEBSITES') if os.getenv('WEBSITES') else args.websites
    urls = {}
    for website in websites.split(','):
        splits = website.split(' ')
        if len(splits) < 2:
            urls[splits[0]] = ''
            continue
        url = splits[0]
        reg_exp = website.lstrip(url).strip()
        urls[url] = reg_exp

    timeout = int(os.getenv('TIMEOUT')) if os.getenv('TIMEOUT') else args.timeout
    period = int(os.getenv('PERIOD')) if os.getenv('PERIOD') else args.period
    kafka = os.getenv('KAFKA') if os.getenv('KAFKA') else args.kafka
    ca = os.getenv('CA') if os.getenv('CA') else args.ca
    cert = os.getenv('CERT') if os.getenv('CERT') else args.cert
    key = os.getenv('KEY') if os.getenv('KEY') else args.key

    # Kafka topics to be created
    # The topics are in the format of 'webactivity-{domain}'
    topics = {}
    for url in urls:
        if not urlparse(url).scheme:
            print(f'Invalid URL {url}, ignored')
            continue
        topics[url] = f'webactiviity-{urlparse(url).netloc}'

    # Create Kafka admin client
    admin = KafkaAdminClient(
        bootstrap_servers=kafka,
        security_protocol="SSL",
        ssl_cafile=ca,
        ssl_certfile=cert,
        ssl_keyfile=key
    )

    # Create Kafka producer
    producer = KafkaProducer(
        bootstrap_servers=kafka,
        security_protocol="SSL",
        ssl_cafile=ca,
        ssl_certfile=cert,
        ssl_keyfile=key
    )

    if os.getenv('PRODUCTION'):
        logger = Logger('tracker', producer, admin)
    else:
        logger = Logger('tracker')

    # Every website's corresponding Kafka topic name is in the following format
    # webactivity-<domain>
    # For example, for https://aiven.io, its topic is: webactivity-aiven.io
    for url in urls.keys():
        # Todo: Make num_partitions and replication_factor for each domain configurable.
        topic_list = [NewTopic(name=topics[url], num_partitions=1, replication_factor=2)]
        try:
            admin.create_topics(new_topics=topic_list, validate_only=False)
        except KafkaErrors.TopicAlreadyExistsError:
            logger.info(f'Topic {topics[url]} already exists')

    # Close admin now since it is not used any more
    admin.close()

    tracker = WebTracker(location, urls, timeout)

    # Forever loop to monitor the websites
    while True:
        stime = time.time()
        for status in tracker.status():
            producer.send(topics[status["url"]], status.serialize())
            logger.info(f'Sending {status}')
            logger.debug(f'Sending {status} with serialized data: {status.serialize()}')
        producer.flush()

        etime = time.time()
        if etime < stime + period:
            logger.info(f'Sleeping for {stime + period - etime} seconds')
            time.sleep(stime + period - etime)


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(
    name='WebsitesAvailability',
    version="v1.1.0",
    author='Liang Hou',
    author_email='eric.hou.liang@gmail.com',
    license='Apache 2.0',
    description='A production-ready Website Availability tracking system',
    url='https://github.com/eric-hou/webavailability',
    setup_requires=['setuptools_scm'],
    packages=find_packages(exclude='tests'),
    scripts=['tracker.py', 'recorder.py'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "msgpack",
        "dnspython",
        "requests",
        "pyOpenSSL",
        "kafka-python",
        "psycopg2",
    ],
    python_requires='>=3.6',
)
